-------------------------------------------------------------------
-- KSI Labs, LLC CENTIPEDE-PCI Mach 3 specific stuff.            --
--                                                               --
-- This is Mach 3 specific stuff. Spindle-related parts.         --
-------------------------------------------------------------------

--------------------------------------------------------------------------
--			LICENSE and Terms of Use.			--
--									--
-- First of all, this is NOT GPL or Free Software. It is open so you	--
-- have the VHDL source code. There is also the source code for the	--
-- Windows Engine for Mach3 CNC software available from KSI Labs, LLC	--
-- website, http://www.ksilabs.com.					--
--									--
-- You are free to use this particular VHDL code, make modifications to	--
-- customize it for your own needs, fix inevitable bugs, port it to 	--
-- other CNC software and do anything else as long as your derivative	--
-- work that is based on this code or uses parts or pieces thereof is	--
-- made to be used with any KSI Labs, LLC products.			--
--									--
-- We do NOT grant you a permission to use any parts of this code in	--
-- any products that are designed for products not manufactured or	--
-- distributed by KSI Labs, LLC. We might give such a permission on	--
-- case-by-case basis but it is not legal to use this code or bits and	--
-- pieces thereof with products not coming from KSI Labs, LLC without	--
-- our explicit written permission.					--
--									--
-- The original copyright notices must be included in all derivative	--
-- works. You do NOT have to ask our permission to make any changes to	--
-- this code and you are not under obligation to give us any of your	--
-- work back as long as you are using it for your own needs.		--
--									--
-- However we do NOT give our permission to distribute your derivative	--
-- work unless all original copyrights are retained, all the conditions	--
-- of this license are met (use only with KSI Labs, LLC products) and	--
-- full source code for such derivative work is available for free to	--
-- everybody without any further conditions including but not limited	--
-- to signing any kinds of NDA. We explicitely deny any permission to	--
-- distribute any derivative work based on this code with closed source	--
-- binary-only fragments or include bits or pieces of this code in such	--
-- products. That essentially means that it is NOT legal to use it with --
-- ANY proprietary HDL megafunctions, megacores, etc. if the resulting	--
-- product is distributed. It IS legal to use whatever you want if the	--
-- resulting product is made for use with a KSI Labs, LLC product AND	--
-- that resulting product is not distributed. Exclusions can be granted --
-- by KSI Labs, LLC on case-by-case basis.				--
--------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.mach3_data_types.all;


entity mach3_spindle is
port (
	sp_din_idx	: in std_logic;
	sp_din_tim	: in std_logic;
	sp_idx_ena	: in std_logic;
	sp_tim_ena	: in std_logic;
	idx_deadtime	: in std_logic_vector(9 downto 0);
	tim_deadtime	: in std_logic_vector(8 downto 0);
	stsp_ena	: in std_logic;
	sp_cnt		: out std_logic_vector(27 downto 0);
	stsp_ctl	: out std_logic_vector(26 downto 0);
	sp_idx_pulse	: out std_logic;
	sp_tim_pulse	: out std_logic;
	sp_stopped	: out std_logic;
	dir_out		: out std_logic;
	step_out	: out std_logic;
	clk520khz	: out std_logic;
	stsp_wr		: in std_logic;
	data_in		: in std_logic_vector(31 downto 0);
	clk		: in std_logic;
	rst		: in std_logic
);
end mach3_spindle;


architecture guts of mach3_spindle is

	-------------
	-- Signals --
	-------------

	signal s_cntr		: unsigned(27 downto 0);
	signal s_cntr_reg	: std_logic_vector(27 downto 0);

	signal time_stb_sig	: std_logic;
	signal fake_stb		: std_logic;

	signal s_idx_sig	: std_logic;
	signal s_idx_prev	: std_logic;
	signal s_tim_sig	: std_logic;
	signal s_tim_prev	: std_logic;
	signal s_idx_pulse	: std_logic;
	signal s_tim_pulse	: std_logic;
	signal s_idx_chg	: std_logic;
	signal s_tim_chg	: std_logic;
	signal dbnc_prediv	: unsigned(8 downto 0);
	signal dbnc_520khz	: std_logic;
	signal dbnc_clk		: std_logic;
	signal idx_dbcntr	: unsigned(9 downto 0);
	signal idx_dbcntr_zero	: std_logic;
	signal idx_dbcntr_dcr	: std_logic;
	signal tim_dbcntr	: unsigned(8 downto 0);
	signal tim_dbcntr_zero	: std_logic;
	signal tim_dbcntr_dcr	: std_logic;
	signal tim_cnt_full	: std_logic;

	signal s_stsp_ctl	: std_logic_vector(26 downto 0);
	signal s_stsp_cntr	: unsigned(24 downto 0);
	signal s_stsp_step	: std_logic;
	signal s_stsp_run	: std_logic;
	signal s_stsp_go	: std_logic;
	signal s_sp_stopped	: std_logic;

begin

	ctl_write_proc : process (rst, clk)
	begin
		if (rst='1') then
			s_stsp_ctl <= (others => '0');
		elsif (rising_edge(clk)) then
			if (stsp_wr='1') then
				s_stsp_ctl <= data_in(26 downto 0);
			end if;
		end if;
	end process;

	s_idx_pulse_proc : process (clk)
	begin
		if (rising_edge(clk)) then
			s_idx_prev <= s_idx_sig;
		end if;
	end process;

	s_tim_pulse_proc : process (clk)
	begin
		if (rising_edge(clk)) then
			s_tim_prev <= s_tim_sig;
		end if;
	end process;

	deadtime_proc : process (rst, clk)
	begin
		if (rst='1') then
			dbnc_prediv <= (others => '0');
			idx_dbcntr <= (others => '0');
			tim_dbcntr <= (others => '0');
		elsif (rising_edge(clk)) then
			if (s_idx_chg='1') then
				idx_dbcntr <= unsigned(idx_deadtime);
			elsif (idx_dbcntr_dcr='1') then
				idx_dbcntr <= idx_dbcntr - 1;
			end if;
			if (s_tim_chg='1') then
				tim_dbcntr <= unsigned(tim_deadtime);
			elsif (tim_dbcntr_dcr='1') then
				tim_dbcntr <= tim_dbcntr - 1;
			end if;
			dbnc_prediv <= dbnc_prediv + 1;
		end if;
	end process;

	spindle_cnt_reg_proc : process (rst, clk)
	begin
		if (rst='1') then
			s_cntr <= (others => '1');
		elsif (rising_edge(clk)) then
			if (time_stb_sig='1' or fake_stb='1') then
				s_cntr_reg <= std_logic_vector(s_cntr);
				s_cntr <= (0 => '1', others => '0');
			elsif (tim_cnt_full='0') then
				s_cntr <= s_cntr + 1;
			end if;
		end if;
	end process;

	stepdir_spindle_proc : process (rst, clk, s_stsp_go)
	begin
		if (rst='1' or s_stsp_go='0') then
			s_stsp_step <= '0';
			s_stsp_cntr <= (others => '0');
		elsif (rising_edge(clk)) then
			if (s_stsp_cntr=(s_stsp_cntr'range => '0')) then
				s_stsp_step <= not s_stsp_step;
				s_stsp_cntr <= unsigned(s_stsp_ctl(24 downto 0));
			else
				s_stsp_cntr <= s_stsp_cntr - 1;
			end if;
		end if;
	end process;

	dbnc_520khz <= '1' when (dbnc_prediv(dbnc_prediv'left-3 downto 0)=(dbnc_prediv'left-3 downto 0 => '1')) else '0';
	dbnc_clk <= '1' when (dbnc_prediv=(dbnc_prediv'range => '1')) else '0';
	idx_dbcntr_zero <= '1' when(idx_dbcntr=(idx_dbcntr'range => '0')) else '0';
	tim_dbcntr_zero <= '1' when(tim_dbcntr=(tim_dbcntr'range => '0')) else '0';
	idx_dbcntr_dcr <= dbnc_clk and not idx_dbcntr_zero;
	tim_dbcntr_dcr <= dbnc_clk and not tim_dbcntr_zero;

	tim_cnt_full <= '1' when (s_cntr=(s_cntr'range => '1')) else '0';

	s_idx_sig <= sp_idx_ena and sp_din_idx;
	s_tim_sig <= sp_tim_ena and sp_din_tim;

	s_idx_pulse <= s_idx_sig and not s_idx_prev and idx_dbcntr_zero;
	s_idx_chg <= (s_idx_sig xor s_idx_prev) and idx_dbcntr_zero;
	sp_idx_pulse <= s_idx_pulse;

	s_tim_pulse <= s_tim_sig and not s_tim_prev and tim_dbcntr_zero;
	s_tim_chg <= (s_tim_sig xor s_tim_prev) and tim_dbcntr_zero;
	sp_tim_pulse <= s_tim_pulse;

	fake_stb <= stsp_wr and data_in(27);
	time_stb_sig <= s_tim_pulse when (sp_tim_ena='1') else s_idx_pulse;

	sp_cnt <= s_cntr_reg;

	s_stsp_run <= s_stsp_ctl(26);
	s_stsp_go <= s_stsp_run and stsp_ena;

	s_sp_stopped <= tim_cnt_full when (sp_idx_ena='1')
				else (stsp_ena and not s_stsp_run);

	dir_out <= s_stsp_ctl(25);
	step_out <= s_stsp_step;
	stsp_ctl <= s_stsp_ctl;
	sp_stopped <= s_sp_stopped;
	
	clk520khz <= dbnc_520khz;

end guts;
