-----------------------------------------------------------------
-- DAC SPI interface. Nothing special, just regular 16-bit SPI --
-- controller. Very simple, only one THR and Idle signal, no   --
-- receive channel (there is nothing to receive from the DAC.) --
-----------------------------------------------------------------

--------------------------------------------------------------------------
--			LICENSE and Terms of Use.			--
--									--
-- First of all, this is NOT GPL or Free Software. It is open so you	--
-- have the VHDL source code. There is also the source code for the	--
-- Windows Engine for Mach3 CNC software available from KSI Labs, LLC	--
-- website, http://www.ksilabs.com.					--
--									--
-- You are free to use this particular VHDL code, make modifications to	--
-- customize it for your own needs, fix inevitable bugs, port it to 	--
-- other CNC software and do anything else as long as your derivative	--
-- work that is based on this code or uses parts or pieces thereof is	--
-- made to be used with any KSI Labs, LLC products.			--
--									--
-- We do NOT grant you a permission to use any parts of this code in	--
-- any products that are designed for products not manufactured or	--
-- distributed by KSI Labs, LLC. We might give such a permission on	--
-- case-by-case basis but it is not legal to use this code or bits and	--
-- pieces thereof with products not coming from KSI Labs, LLC without	--
-- our explicit written permission.					--
--									--
-- The original copyright notices must be included in all derivative	--
-- works. You do NOT have to ask our permission to make any changes to	--
-- this code and you are not under obligation to give us any of your	--
-- work back as long as you are using it for your own needs.		--
--									--
-- However we do NOT give our permission to distribute your derivative	--
-- work unless all original copyrights are retained, all the conditions	--
-- of this license are met (use only with KSI Labs, LLC products) and	--
-- full source code for such derivative work is available for free to	--
-- everybody without any further conditions including but not limited	--
-- to signing any kinds of NDA. We explicitely deny any permission to	--
-- distribute any derivative work based on this code with closed source	--
-- binary-only fragments or include bits or pieces of this code in such	--
-- products. That essentially means that it is NOT legal to use it with --
-- ANY proprietary HDL megafunctions, megacores, etc. if the resulting	--
-- product is distributed. It IS legal to use whatever you want if the	--
-- resulting product is made for use with a KSI Labs, LLC product AND	--
-- that resulting product is not distributed. Exclusions can be granted --
-- by KSI Labs, LLC on case-by-case basis.				--
--------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity tlv5617a_dac is
generic (
	clkdiv_val	: unsigned(5 downto 0) := b"10_0010"
);
port (
	dac_idle	: out std_logic;
	ss_pin		: out std_logic;
	sclk_pin	: out std_logic;
	mosi_pin	: out std_logic;
	dac_thr		: in std_logic_vector(15 downto 0);
	dac_sel		: in std_logic;
	wr		: in std_logic;
	clk		: in std_logic;
	rst		: in std_logic
);
end tlv5617a_dac;

architecture guts of tlv5617a_dac is

	-------------
	-- Signals --
	-------------

	type dac_fsm_state is (DAC_RDY, DAC_START, DAC_LATCH, DAC_SHIFT);
	signal fsm_cur		: dac_fsm_state;
	signal fsm_nxt		: dac_fsm_state;
	
	signal thr		: std_logic_vector(15 downto 0);
	signal clkdiv		: unsigned(5 downto 0);
	signal clkdiv_load	: std_logic;
	signal state_change	: std_logic;
	signal bit_cntr		: unsigned(3 downto 0);
	signal do_load		: std_logic;
	signal do_shift		: std_logic;
	signal dac_write	: std_logic;
	signal dac_idle_sig	: std_logic;

begin

	--------------------------------------
	-- Top half of the FSM (sequential) --
	--------------------------------------
	fsm_top : process (rst, clk)
	begin
		if (rst='1') then
			fsm_cur <= DAC_RDY;
			clkdiv <= clkdiv_val;
			bit_cntr <= (others => '0');
		elsif (rising_edge(clk)) then
			if (do_load='1') then
				thr <= dac_thr;
				bit_cntr <= (others => '0');
			elsif (do_shift='1') then
				thr <= thr(14 downto 0) & '0';
				bit_cntr <= bit_cntr + 1;
			end if;
			if (clkdiv_load='1') then
				clkdiv <= clkdiv_val;
			else
				clkdiv <= clkdiv - 1;
			end if;
			fsm_cur <= fsm_nxt;
		end if;
	end process;
	
	--------------------------------------------
	-- Bottom half of the FSM (combinatorial) --
	--------------------------------------------
	fsm_bh : process (fsm_cur, state_change, dac_write, bit_cntr)
	begin
		case fsm_cur is
			when DAC_RDY =>
				if (dac_write='1') then
					fsm_nxt <= DAC_START;
				else
					fsm_nxt <= DAC_RDY;
				end if;
			when DAC_START =>
				if (state_change='1') then
					fsm_nxt <= DAC_LATCH;
				else
					fsm_nxt <= DAC_START;
				end if;
			when DAC_LATCH =>
				if (state_change='1') then
					if (bit_cntr="1111") then
						fsm_nxt <= DAC_RDY;
					else
						fsm_nxt <= DAC_SHIFT;
					end if;
				else
					fsm_nxt <= DAC_LATCH;
				end if;
			when DAC_SHIFT =>
				if (state_change='1') then
					fsm_nxt <= DAC_LATCH;
				else
					fsm_nxt <= DAC_SHIFT;
				end if;
		end case;
	end process;

	dac_write <= dac_sel and wr;

	dac_idle_sig <= '1' when (fsm_cur=DAC_RDY) else '0';

	state_change <= '1' when (clkdiv=(clkdiv'range => '0')) else '0';

	do_load <= dac_write and dac_idle_sig;

	do_shift <= '1' when (fsm_nxt=DAC_SHIFT and state_change='1') else '0';

	clkdiv_load <= state_change or dac_idle_sig;

	dac_idle <= dac_idle_sig;
	
	ss_pin <= '1' when (fsm_cur=DAC_RDY) else '0';
	sclk_pin <= '0' when (fsm_cur=DAC_LATCH) else '1';
	mosi_pin <= thr(15);

end guts;
